import logging
import os
import optilibre.helpers as helpers

from optilibre import models as models, enums as enums
from optilibre.helpers import path_exist, get_path_as_escaped_str
from pathlib import Path


def cmd_src_on_success_failure(config, subdir_relative: Path, on_success_failure: 'success, failure'):
    """
    Return shell cmd to execute on success or failure for the src file.
    :param config:
    :param subdir_relative:
    :param on_success_failure:
    :return:
    """
    src_on_keyword = 'src_on_' + on_success_failure
    src_on_keyword_dest = src_on_keyword + '_dest'
    if src_on_keyword not in config:
        logging.info("%s is not defined, default back to 'keep'" % src_on_keyword)
        config[src_on_keyword] = 'keep'

    on_success_failure = config[src_on_keyword]
    cmd_on_success_failure = ""
    if on_success_failure == 'keep':
        pass
    elif on_success_failure == 'move':
        if src_on_keyword_dest not in config:
            logging.warning("%s was defined as 'move' but not %s was defined. Default back to 'keep'." %
                            (src_on_keyword, src_on_keyword_dest))

        dest = config[src_on_keyword_dest]
        if subdir_relative:
            dest = os.path.join(dest, subdir_relative)
            helpers.ensure_folder_exists(dest)
        if not path_exist(dest):
            logging.warning("Default back to keep for %s" % on_success_failure)
        else:
            cmd_on_success_failure = "mv {0} " + get_path_as_escaped_str(path=dest) + "/{1}"
            logging.debug("CMD on success failure (move) {}".format(cmd_on_success_failure))
    elif on_success_failure == 'delete':
        cmd_on_success_failure = "rm {0}"

    return cmd_on_success_failure


def do_src_on_done(exit_code: int, file, local_config, subdir_relative: Path):
    """
    Launch shell cmd to execute on success or failure, to move, delete ... src file.
    :param subdir_relative: (if subdir) subdir where the file was located (relative to in_path)
    :param local_config:
    :param exit_code:
    :param file:
    :return:
    """

    if exit_code == 0:
        cmd = cmd_src_on_success_failure(config=local_config, subdir_relative=subdir_relative, on_success_failure='success')
        logging.debug("Exit code was successful.")
    else:
        cmd = cmd_src_on_success_failure(config=local_config, subdir_relative=subdir_relative, on_success_failure='failure')
        logging.debug("Exit code was not successful %s" % str(exit_code))

    cmd = cmd.format(get_path_as_escaped_str(path=file), get_path_as_escaped_str(path=os.path.basename(file)))
    logging.debug("Execute post convert (empty if 'keep'): %s" % cmd)
    os.system(cmd)

    return None


def build_cmdline_video(folder: models.VideoFolder) -> str:
    """
    Build cmdline options
    :param folder:
    :return: cmdline (without in and out files)
    """

    codec_video = folder.get_video_codec()
    local_config = folder.get_config()

    # Set default options if not already set
    if "map_metadata" not in local_config['meta']:
        local_config['meta']['map_metadata'] = 0
    if "n" not in local_config['meta'] or "y" not in local_config['meta']:
        local_config['meta']['n'] = ''
    if codec_video == enums.VideoCodec.libx265 and "dst_range" not in local_config[codec_video.name]:
        # enable full color range for x265 per default
        local_config[codec_video.name]['dst_range'] = 1

    # Build cmdline
    cmdline_meta = ""
    if 'meta' in local_config:
        for opt in local_config['meta'].items():
            cmdline_meta += " -" + str(opt[0]) + " " + str(opt[1])

    cmdline_audio = ""
    if local_config['audio']['enabled']:
        cmdline_audio = "-c:a " + str(folder.get_audio_codec())

    cmdline_video = ""
    if local_config['video']['enabled']:
        cmdline_video = "-c:v " + codec_video.name
        for opt in local_config[local_config['video']['codec']].items():
            cmdline_video += " -" + str(opt[0]) + " " + str(opt[1])

    return "ffmpeg -i " + "'{}'" + cmdline_meta + " " + cmdline_audio + " " + cmdline_video + " " + "'{}'"



import toml
import logging
import optilibre.models as models
import optilibre.helpers as helpers


def parse_config_file(config_file_path):
    """
    Parse the config file to build and store configuration in a global_config object.
    @param config_file_path: path to optilibre.conf
    """
    logging.info("Parsing " + config_file_path)
    logging.debug("Any error will be fatal.")
    with open(config_file_path, 'r') as f:
        config = toml.load(f)

    version = helpers.get_conf_version(conf=config)
    global_config = models.GlobalConfig(version=version)

    # Check which type are defined
    types_ = []
    if "image" in config['convert']:
        types_.append("image")
    if "video" in config['convert']:
        types_.append("video")

    # Build image and video "folders"
    for type_ in types_:
        logging.debug("Folders {}".format(config['convert'][type_]))
        for folder_name in config['convert'][type_]:
            logging.info("Parsing {}".format(folder_name))
            folder_config = helpers.get_local_config(conf=config['convert'][type_][folder_name])
            enabled = helpers.is_config_enabled(config=folder_config)
            if not enabled:
                logging.info("The folder {} is disabled. Continue with the next folder.".format(folder_name))
                continue

            try:
                if type_ == "image":
                    folder_name = models.ImageFolder(name=folder_name, enabled=enabled, config_as_json=folder_config["optiimage"])
                    global_config.image_folder_list.append(folder_name)
                elif type_ == "video":
                    folder_name = models.VideoFolder(name=folder_name, enabled=enabled, config_as_json=folder_config["optivideo"])
                    global_config.video_folder_list.append(folder_name)
                else:
                    raise TypeError("{} unknown.".format(type_))
            except Exception:
                logging.debug("folder_config: {}".format(folder_config))
                # Forward current stack trace
                raise

    return global_config







import filetype
import logging
import os
import re
import optilibre
import optilibre.core
import optilibre.enums as enums
import optilibre.helpers as helpers
import optilibre.models as models
import optilibre.utils.parsers as parsers
from multiprocessing import Pool
from pathlib import Path

from optilibre.core import build_cmdline_video


def convert_img(file: Path, shell_cmdline: str, local_config: dict, subdir_relative: Path):
    logging.debug("Converting " + str(file))

    # TODO move file extension / output
    logging.info(shell_cmdline)
    exit_code = os.WEXITSTATUS(os.system(shell_cmdline))
    logging.debug("Exit code: %s" % exit_code)
    optilibre.core.do_src_on_done(exit_code=exit_code, file=file, local_config=local_config, subdir_relative=subdir_relative)


def process_image_folder(folder: models.ImageFolder, subdir: Path = None):
    """
    This function recursively process an "optilibre" folder and its subdir.
    For each subdir encountered, call itself with subdir as arg.
    For the subdir in argument, create the subdir in the dest folders.
    """
    in_path = folder.get_in_path()

    subdir_msg = ""
    if subdir:
        subdir_msg = "(subdir: {})".format(subdir.relative_to(in_path))
    logging.info("Treating folder: {} {}".format(folder.get_name(), subdir_msg))

    local_config = folder.get_config()
    out_path = folder.get_out_path()
    codec = folder.get_codec()

    relative_subdir = None
    if subdir:
        # If we are working with a sub dir. Append it to out_path.
        relative_subdir = subdir.relative_to(in_path)  # for src_on_done
        out_path = out_path.joinpath(relative_subdir)  # get subdir from absolute path and join it with out_path
        in_path = subdir
        helpers.ensure_folder_exists(path=out_path)

    # Set default options for JPEGoptim
    if codec == enums.ImageCodec.jpeg:
        # Remove dest if provided (otherwise subdir won't work because of arg as reference)
        local_config[codec.name.lower()].pop('d', None)
        local_config[codec.name.lower()].pop('dest', None)
        #if "d" not in local_config[codec.name.lower()] and "dest" not in local_config[codec.name.lower()]:
        local_config[codec.name.lower()]['d'] = helpers.get_path_as_escaped_str(out_path)

    # Build cmdline options
    cmdline_codec_opts = ""
    if codec.name.lower() in local_config:
        for opt in local_config[codec.name.lower()].items():
            cmdline_codec_opts += " -" + str(opt[0]) + " " + str(opt[1])
    else:
        logging.debug("No config for %s found. Using default codec values." % codec.name)

    arg_list = []  # [(file1, shell1, local_config, subdir), (...)] we split arg_list to processes
    for file in Path(in_path).glob('*'):
        if file.is_file():
            kind = filetype.guess_mime(str(file))
            if bool(re.match("image/+", str(kind))):
                logging.debug("Building cmd line for: %s" % str(file))
                file_mime = filetype.guess_mime(str(file))

                if file.suffix not in optilibre.supported_img_ext or not bool(re.match('image/+', str(file_mime))):
                    # TODO Jpeg-XL can convert non jpeg format.
                    logging.warning("Filetype: %s not supported: for file: %s. Won't convert it." % (str(file_mime), str(file)))
                    # CONTINUE
                    continue

                # JPEG jpegoptim
                if codec == enums.ImageCodec.jpeg:
                    encoder_cmdline = "jpegoptim"
                    out_file = ""

                # JPEG XL
                elif codec == enums.ImageCodec.jpegxl:
                    # TODO implement recursive
                    encoder_cmdline = "cjxl"
                    out_file = os.path.join(out_path, os.path.splitext(os.path.basename(file))[0] + ".jxl")
                else:
                    logging.error("Codec type not supported %s" % str(codec))
                    # CONTINUE
                    continue

                shell_cmdline = "{encoder} {in_file} {options} {outfile}".format(
                    encoder=encoder_cmdline, in_file=helpers.get_path_as_escaped_str(file), options=cmdline_codec_opts, outfile=out_file)

                arg_list.append((file, shell_cmdline, local_config, relative_subdir))
            else:
                # Changing suffix
                logging.debug("Filetype: %s for file: %s isn't an image file." % (str(kind), str(file)))
        elif file.is_dir():
            logging.debug("{} is a directory. Recursively entering it.".format(str(file)))
            process_image_folder(folder=folder, subdir=Path(file))
        else:
            logging.debug("{} is not a file neither a dir. Ignoring it.".format(str(file)))

    # multiprocess call
    if arg_list:
        with Pool() as p:
            p.starmap(convert_img, arg_list)


def process_video_folder(folder: models.VideoFolder, subdir: Path = None):
    """
    This function recursively process an "optilibre" folder and its subdir.
    For each subdir encountered, call itself with subdir as arg.
    For the subdir in argument, create the subdir in the dest folders.
    """
    in_path = folder.get_in_path()

    subdir_msg = ""
    if subdir:
        subdir_msg = "(subdir: {})".format(subdir.relative_to(in_path))
    logging.info("Treating folder: {} {}".format(folder.get_name(), subdir_msg))

    local_config = folder.get_config()
    out_path = folder.get_out_path()

    relative_subdir = None
    if subdir:
        # If we are working with a sub dir. Append it to out_path.
        relative_subdir = subdir.relative_to(in_path)  # for src_on_done
        out_path = out_path.joinpath(relative_subdir)  # get subdir from absolute path and join it with out_path
        in_path = subdir
        helpers.ensure_folder_exists(path=out_path)

    cmdline = build_cmdline_video(folder=folder)

    for file in Path(in_path).glob('*'):  # Path(Path() = Path()
        if file.is_file():
            kind = filetype.guess_mime(str(file))

            if bool(re.match("video/+", str(kind))):
                logging.debug("Processing " + str(file))

                if file.suffix in optilibre.supported_video_ext:
                    out_file = os.path.join(out_path, os.path.basename(
                        file.with_suffix('.' + str(folder.get_format_container()))
                    ))

                    ffmpeg_cmdline = cmdline.format(file, out_file)

                    logging.info(ffmpeg_cmdline)
                    exit_code = os.WEXITSTATUS(os.system(ffmpeg_cmdline))
                    optilibre.core.do_src_on_done(exit_code=exit_code, file=file, local_config=local_config, subdir_relative=relative_subdir)

                else:
                    logging.info("Filetype {} for {} is not (yet) supported.".format(str(file.suffix), str(file)))
            else:
                # Changing suffix
                logging.debug("Filetype: %s for file: %s isn't a video file. Ignoring it." % (str(kind), str(file)))
        elif file.is_dir():
            logging.debug("{} is a directory. Recursively entering it.".format(str(file)))
            process_video_folder(folder=folder, subdir=Path(file))
        else:
            logging.debug("{} is not a file. Ignoring it.".format(str(file)))


def main(config_file):
    global_config = parsers.parse_config_file(config_file_path=config_file)

    logging.info("Processing image files.")
    image_folder: models.ImageFolder
    for image_folder in global_config.get_image_folders():
        process_image_folder(folder=image_folder)

    logging.info("Processing video files.")
    video_folder: models.VideoFolder
    for video_folder in global_config.get_video_folders():
        process_video_folder(folder=video_folder)




import coloredlogs
import logging
import importlib.metadata

__version__ = "Unknown"
try:
    __version__ = importlib.metadata.version('optilibre')
except importlib.metadata.PackageNotFoundError:
    # package is not installed
    logging.warning("Optilibre (unknown version). Is the package installed ?")

coloredlogs.install(level=logging.DEBUG)

supported_config_version_min = 1.0
supported_img_ext = [".jpeg", ".jpg", ".jpegxl", ".jxl"]
supported_video_ext = [".mp4", ".mkv", ".webm"]



import logging
import os
import shutil
import pathlib
import typing

import toml
from optilibre import supported_config_version_min


def ensure_folder_exists(path: pathlib.Path):
    try:
        if not os.path.exists(path):
            os.makedirs(path)
    except PermissionError:
        logging.exception("Can't create path {}. No Permission.".format(str(path)))


def get_conf_version(conf):
    """return main config file version"""
    try:
        version = conf['optilibre']['version']
    except KeyError:
        logging.warning("Version of the main config file could not be found. Assuming we are at least at %s. " % str(supported_config_version_min))
        version = supported_config_version_min

    return version


def get_local_conf_version(conf):
    """return local (sub) config file versiob"""
    try:
        if 'optiimage' in conf:
            version = conf['optiimage']['version']
        elif 'optivideo' in conf:
            version = conf['optivideo']['version']
        else:
            raise KeyError

    except KeyError:
        logging.warning("Version of the local config file could not be found. Assuming we are at least at %s. " % str(supported_config_version_min))
        logging.debug(conf)
        version = supported_config_version_min

    return version


def get_escaped_path(path: typing.Union[str, pathlib.Path]) -> pathlib.Path:
    """
    Return a Path from provided path (it helps convert str with escape char (\\) to a valid Path)
    """
    if type(path) is str and '\\' in path:
        return pathlib.Path(path.replace('\\', ''))
    else:
        return pathlib.Path(path)


def get_path_as_escaped_str(path: pathlib.Path) -> str:
    return str(path).replace(" ", "\\ ")


def check_version(conf_version):
    if conf_version < supported_config_version_min:
        raise ValueError("Config version ({}) mismatched minimum supported version ({})".format(
            conf_version, supported_config_version_min
        ))
    else:
        logging.debug("Config file version ({}) is supported".format(conf_version))


def check_local_version(local_conf):
    conf_version = get_local_conf_version(conf=local_conf)
    if conf_version < supported_config_version_min:
        raise ValueError("Config version ({}) mismatched minimum supported version ({})".format(
            conf_version, supported_config_version_min
        ))
    else:
        logging.debug("Local config file version ({}) is supported".format(conf_version))


def is_config_enabled(config):
    if "enabled" in config:
        return config["enabled"]
    else:
        logging.debug("No enabled line found. Assuming it is.")
        return True


def is_tool(name):
    """Check whether `name` is on PATH and marked as executable."""

    return shutil.which(name) is not None


def path_exist(path):
    if not os.path.exists(path) or not os.path.isdir(path):
        logging.error("Path (%s) doesn't exist or is not a directory." % path)
        # RETURN
        return False
    else:
        # RETURN
        return True


def _get_or_raise_path(folder_config, key: str) -> pathlib.Path:
    if key in folder_config:
        path = get_escaped_path(folder_config[key])
    else:
        logging.error("_get_or_raise_path: there is no key named {} in folder_config: {}".format(key, folder_config))
        raise FileNotFoundError("%s is not in configuration." % key)
    if not path_exist(path):
        logging.error("_get_or_raise_path: {} is not a valid path. folder_config: {}".format(path, folder_config))
        raise FileNotFoundError("%s not a valid path." % path)
    else:
        return path


def get_or_raise_in_path(folder_config):
    return _get_or_raise_path(folder_config=folder_config, key="in_path")


def get_or_raise_out_path(folder_config):
    return _get_or_raise_path(folder_config=folder_config, key="out_path")


def get_local_config(conf) -> dict:
    """
    Open and merge local (folder specific) config file if configured
    :param conf: main config file
    :return: local config file
    """
    local_config = conf

    if 'config_file' in conf:
        local_conf_file = os.path.join(conf['config_file'])
    elif 'path' in conf:
        logging.warning("Path= is not supported anymore. Migrate to config_file")
        raise KeyError("Path= is not supported anymore. Migrate to config_file=")
    else:
        logging.debug("No local config file given. Assuming the configuration is in main config file.")
        local_conf_file = None

    if local_conf_file is not None:
        try:
            with open(local_conf_file, 'r') as f:
                file_data = toml.load(f)

            # Add content from file
            for type_ in ("optiimage", "optivideo"):
                if type_ in file_data:
                    local_config[type_] = file_data[type_]

        except OSError or FileNotFoundError:
            logging.exception("Could not find {} conf file for {}.".format(local_conf_file, local_config))
            local_config = None
            exit(13)  # EXIT

    logging.debug("get_local_config: Current configuration: {}".format(
        local_config
    ))
    check_local_version(local_conf=local_config)

    return local_config



import logging
import optilibre.cli as cli

from optilibre import __version__


def main():
    logging.info("Optilibre v{}".format(__version__))
    cli.cli()


if __name__ == '__main__':
    main()



from enum import Enum


class ImageCodec(Enum):
    jpeg = "jpeg"
    jpegxl = "jpegxl"

    def __str__(self):
        return self.value

    @classmethod
    def has_key(cls, name):
        return name in cls.__members__


class AudioCodec(Enum):
    aac = "aac"
    ogg = "ogg"

    def __str__(self):
        return self.value

    @classmethod
    def has_key(cls, name):
        return name in cls.__members__


class VideoCodec(Enum):
    libx264 = "libx264"
    libx265 = "libx265"

    def __str__(self):
        return self.value

    @classmethod
    def has_key(cls, name):
        return name in cls.__members__


class VideoContainer(Enum):
    avi = "avi"
    mkv = "mkv"
    mp4 = "mp4"

    def __str__(self):
        return self.value

    @classmethod
    def has_key(cls, name):
        return name in cls.__members__



import json
import logging
import optilibre.enums as enums
import optilibre.helpers as helpers
from pathlib import Path


class Folder:
    def __init__(self, name, enabled, config_as_json):
        self.name: str = name
        self.enabled: bool = enabled
        self.config: json = config_as_json  # json located in 'optiimage' or 'optivideo'

        self.in_path: Path = helpers.get_or_raise_in_path(folder_config=self.get_config())
        self.out_path: Path = helpers.get_or_raise_out_path(folder_config=self.get_config())

    def get_config(self) -> dict:
        return self.config

    def get_name(self) -> str:
        return self.name

    def get_in_path(self) -> Path:
        return self.in_path

    def get_out_path(self) -> Path:
        return self.out_path


class ImageFolder(Folder):
    def get_codec(self):
        codec_str = self.get_config()["codec"]
        try:
            codec = enums.ImageCodec[codec_str.lower()]
        except KeyError:
            logging.error("%s is not a supported video codec." % codec_str)
            raise KeyError

        return codec

    @staticmethod
    def get_opti_type():
        return "optiimage"


class VideoFolder(Folder):
    def get_audio_codec(self) -> enums.AudioCodec:
        codec_str = self.get_config()['audio']['codec']
        try:
            codec = enums.AudioCodec(codec_str.lower())
        except KeyError:
            logging.error("%s is not a supported video codec." % codec_str)
            raise KeyError

        return codec

    def get_video_codec(self) -> enums.VideoCodec:
        codec_str = self.get_config()['video']['codec']
        try:
            codec = enums.VideoCodec[codec_str.lower()]
        except KeyError:
            logging.error("%s is not a supported video codec." % codec_str)
            raise KeyError

        return codec

    def get_format_container(self) -> enums.VideoContainer:
        if 'container' in (config := self.get_config()):
            container = config['container']
        else:
            # if no container is defined or is invalid
            video_codec = self.get_video_codec()
            if video_codec in [enums.VideoCodec.libx264, enums.VideoCodec.libx265]:
                container = enums.VideoContainer.mp4
            else:
                logging.error("Could not find a suitable container for codec {}".format(video_codec))
                container = None

        try:
            container = enums.VideoContainer(container)
        except ValueError:
            logging.error("Desired container format ({}) is not supported.")
            raise

        return container

    @staticmethod
    def get_opti_type():
        return "optivideo"


class GlobalConfig:
    def __init__(self, version):
        self.version = version
        self.image_folder_list = []
        self.video_folder_list = []

        helpers.check_version(conf_version=self.get_version())

    def get_image_folders(self) -> [ImageFolder]:
        return self.image_folder_list

    def get_video_folders(self) -> [VideoFolder]:
        return self.video_folder_list

    def get_version(self) -> float:
        return self.version
